/* $Id: pci_register.h,v 1.1 1999/06/18 05:50:51 vallon Exp $ */

#ifndef _PPC_POWERMAC_PCI_PCI_REGISTER_H_
#define _PPC_POWERMAC_PCI_PCI_REGISTER_H_

int pci_register(void);
/* pci_register registers PCI drivers.  Returns 0 on success.  */

#endif

