/* conf.h -- Dave's Magic Configuration File
 *
 *
 */

#ifndef __conf_h__

#if 1
#define PB1400_EXPBAY
#endif

#if 0
#define CONSOLE_CACHE
#endif

#if 0
#define IDE_ZIP_HACK
#endif

#if 1
#define ORIG_PMAP
#endif

#if 0
#define OLD_PROBE
#endif

#define __conf_h__
#endif
