/*
 * Copyright 1991-1998 by Open Software Foundation, Inc. 
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 * 
 */
/*
 * MkLinux
 */
/*	Generated by configure. Don't edit!!!!	*/

#define KERNEL_SOUNDCARD
#define EXCLUDE_PAS
#undef  EXCLUDE_SB
#define EXCLUDE_ADLIB
#define EXCLUDE_GUS
#define EXCLUDE_MPU401
#define EXCLUDE_UART6850
#define EXCLUDE_PSS
#define EXCLUDE_GUS16
#define EXCLUDE_GUSMAX
#define EXCLUDE_MSS
#undef  EXCLUDE_SBPRO
#undef  EXCLUDE_SB16
#undef  EXCLUDE_AUDIO
#undef  EXCLUDE_MIDI
#undef  EXCLUDE_YM3812
#undef  EXCLUDE_SEQUENCER

#define DSP_BUFFSIZE 32768
#define SELECTED_SOUND_OPTIONS	0x00035c02
#define SOUND_VERSION_STRING "2.90-2"
#define SOUND_CONFIG_DATE "Fri Dec 1 18:12:35 EST 1995"
#define SOUND_CONFIG_BY "naoki"
#define SOUND_CONFIG_HOST "salt"
#define SOUND_CONFIG_DOMAIN "osf.org"
